"""Configuration for connecting to the AMaaS API."""
from __future__ import absolute_import
from collections import namedtuple
from configparser import ConfigParser, Error as ConfigParserError
import os
import re

from .exceptions import ConfigurationError, MissingConfigurationError


ENDPOINTS = {
    'asset_managers': 'assetmanager',
    'assets': 'asset',
    'books': 'book',
    'corporate_actions': 'corporateaction',
    'fundamentals': 'fundamental',
    'market_data': 'marketdata',
    'monitor': 'monitor',
    'parties': 'party',
    'transactions': 'transaction'
}


arn_re = re.compile(
    r'arn:(?P<partition>[\w-]+):(?P<service>[\w-]+):(?P<region>[\w-]*):(?P<account_id>\d*):'
    r'((?P<resourcetype>[\w-]*)[/:])?'
    r'(?P<resource>\S*)'
)


APIConfig = namedtuple('APIConfig', ['api_url'])


AuthConfig = namedtuple('AuthConfig', [
    'username', 'password',
    'cognito_pool_id', 'cognito_region', 'cognito_client_id',
])


class ConfigFactory(object):
    """Factory for building config object."""

    known_api_configurations = {
        'prod': APIConfig(api_url='https://api.amaas.com/sg1.0/'),
        'staging': APIConfig(api_url='https://api-staging.amaas.com/v1.0/'),
        'dev': APIConfig(api_url='https://api-dev.amaas.com/v1.0/'),
        'local': APIConfig(api_url='http://localhost:8000/'),
    }

    known_auth_configurations = {
        'prod': dict(
            cognito_pool_id='ap-southeast-1_0LilJdUR3',
            cognito_region='ap-southeast-1',
            cognito_client_id='7b7kt883veb7rr2toolj1ukp6j',
        ),
        'staging': dict(
            cognito_pool_id='ap-southeast-1_De6j7TWIB',
            cognito_region='ap-southeast-1',
            cognito_client_id='2qk35mhjjpk165vssuqhqoi1lk',
        ),
        'dev': dict(
            cognito_pool_id='ap-southeast-1_De6j7TWIB',
            cognito_region='ap-southeast-1',
            cognito_client_id='2qk35mhjjpk165vssuqhqoi1lk',
        ),
        'local': dict(
            cognito_pool_id='ap-southeast-1_De6j7TWIB',
            cognito_region='ap-southeast-1',
            cognito_client_id='2qk35mhjjpk165vssuqhqoi1lk',
        ),
    }

    def __init__(self, config_filepath=None):
        """Create new config factory."""
        # Read config file
        config_files = [config_filepath] if config_filepath else []
        config_files.extend([
            '.amaas.cfg',
            os.path.expanduser(os.path.join('~', '.amaas.cfg')),
            os.path.join('', 'etc', 'amaas.cfg'),
        ])
        parser = ConfigParser()
        parser.read(config_files)
        self.file_config = parser

    def lookup(self, section, name):
        """Lookup config value."""
        value = os.environ.get('AMAAS_{}'.format(name.upper()))
        if value:
            return value

        try:
            value = self.file_config.get(section, name)
        except ConfigParserError:
            pass
        else:
            if value:
                return value

        raise MissingConfigurationError(section, name)

    def api_config(self, stage=None):
        """Create api config based on stage."""
        if stage in self.known_api_configurations:
            return self.known_api_configurations[stage]

        if not stage:
            section = 'stages.live'
            api_url = 'https://api.amaas.com/'
        else:
            section = 'stages.{}'.format(stage)
            try:
                api_url = self.lookup(section, 'api_url')
            except ConfigurationError:
                api_url = 'https://api-{}.dev.amaas.com/'.format(stage)

        return APIConfig(api_url)

    def auth_config(self, stage=None):
        """Create auth config based on stage."""
        if stage:
            section = 'stages.{}'.format(stage)
        else:
            section = 'stages.live'

        try:
            username = self.lookup(section, 'username')
            password = self.lookup(section, 'password')
        except MissingConfigurationError:
            username = self.lookup('auth', 'username')
            password = self.lookup('auth', 'password')

        if stage in self.known_auth_configurations:
            return AuthConfig(
                username, password,
                **self.known_auth_configurations[stage]
            )

        try:
            cognito_pool_id = self.lookup(section, 'cognito_pool_id')
            cognito_region = self.lookup(section, 'cognito_region')
        except MissingConfigurationError:
            m = arn_re.match(self.lookup(section, 'cognito_pool'))
            if not m:
                raise ConfigurationError('Cognito Pool value must be an ARN')

            cognito_pool_id = m.group('resource')
            cognito_region = m.group('region')

        cognito_client_id = self.lookup(section, 'cognito_client_id')
        return AuthConfig(
            username, password,
            cognito_pool_id, cognito_region, cognito_client_id,
        )


_global_config = None


def get_factory():
    """Get global config factory."""
    global _global_config
    if not _global_config:
        _global_config = ConfigFactory()

    return _global_config



from __future__ import absolute_import, division, print_function, unicode_literals

from collections import Iterable
import json
import logging

from amaascore.assets.utils import json_to_asset
from amaascore.core.interface import Interface
from amaascore.core.amaas_model import json_handler


class AssetsInterface(Interface):

    def __init__(self, environment=None, endpoint=None, logger=None, username=None,
                       password=None, session_token=None):
        self.logger = logger or logging.getLogger(__name__)
        super(AssetsInterface, self).__init__(endpoint=endpoint, endpoint_type='assets', session_token=session_token,
                                              environment=environment, username=username, password=password)

    def new(self, asset):
        self.logger.info('New Asset - Asset Manager: %s - Asset ID: %s', asset.asset_manager_id, asset.asset_id)
        url = '%s/assets/%s' % (self.endpoint, asset.asset_manager_id)
        response = self.session.post(url, json=asset.to_interface())
        if response.ok:
            self.logger.info('Successfully Created Asset - Asset Manager: %s - Asset ID: %s', asset.asset_manager_id,
                             asset.asset_id)
            asset = json_to_asset(response.json())
            return asset
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def create_many(self, assets):
        if not assets or not isinstance(assets, list):
            raise ValueError('Invalid argument. Argument must be a non-empty list.')

        self.logger.info('New Assets - Asset Manager: %s', assets[0].asset_manager_id)
        url = '%s/assets/%s' % (self.endpoint, assets[0].asset_manager_id)
        json_body = [asset.to_interface() for asset in assets]
        response = self.session.post(url, json=json_body)
        if response.ok:
            self.logger.info('Successfully Created Assets - Asset Manager: %s', assets[0].asset_manager_id)
            assets = [asset for asset in response.json()]
            return assets
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def upsert(self, asset):
        ''' upsert only support upserting one asset at a time'''
        self.logger.info('Upsert Asset - Asset Manager: %s - Asset ID: %s', asset.asset_manager_id, asset.asset_id)
        url = '%s/assets/%s' % (self.endpoint, asset.asset_manager_id)
        response = self.session.post(url, json=asset.to_interface(), params={'upsert': True})
        if response.ok:
            self.logger.info('Successfully Upserted Asset - Asset Manager: %s - Asset ID: %s', asset.asset_manager_id,
                             asset.asset_id)
            asset = json_to_asset(response.json())
            return asset
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def amend(self, asset):
        self.logger.info('Amend Asset - Asset Manager: %s - Asset ID: %s', asset.asset_manager_id, asset.asset_id)
        url = '%s/assets/%s/%s' % (self.endpoint, asset.asset_manager_id, asset.asset_id)
        response = self.session.put(url, json=asset.to_interface())
        if response.ok:
            self.logger.info('Successfully Amended Asset - Asset Manager: %s - Asset ID: %s', asset.asset_manager_id,
                             asset.asset_id)
            asset = json_to_asset(response.json())
            return asset
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def partial(self, asset_manager_id, asset_id, updates):
        self.logger.info('Partial Amend Asset - Asset Manager: %s - Asset ID: %s', asset_manager_id,
                         asset_id)
        url = '%s/assets/%s/%s' % (self.endpoint, asset_manager_id, asset_id)
        # Setting handler ourselves so we can be sure Decimals work
        response = self.session.patch(url, data=json.dumps(updates, default=json_handler), headers=self.json_header)
        if response.ok:
            asset = json_to_asset(response.json())
            return asset
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve(self, asset_manager_id, asset_id, version=None):
        self.logger.info('Retrieve Asset - Asset Manager: %s - Asset ID: %s', asset_manager_id, asset_id)
        url = '%s/assets/%s/%s' % (self.endpoint, asset_manager_id, asset_id)
        if version:
            url += '?version=%d' % int(version)
        response = self.session.get(url)
        if response.ok:
            self.logger.info('Successfully Retrieved Asset - Asset Manager: %s - Asset ID: %s', asset_manager_id,
                             asset_id)
            return json_to_asset(response.json())
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def deactivate(self, asset_manager_id, asset_id):
        self.logger.info('Deactivate Asset - Asset Manager: %s - Asset ID: %s', asset_manager_id, asset_id)
        url = '%s/assets/%s/%s' % (self.endpoint, asset_manager_id, asset_id)
        json = {'asset_status': 'Inactive'}
        response = self.session.patch(url, json=json)
        if response.ok:
            self.logger.info('Successfully Deactivated Asset - Asset Manager: %s - Asset ID: %s', asset_manager_id,
                             asset_id)
            return json_to_asset(response.json())
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def loose_search(self, asset_manager_id, query='', **kwargs):
        """
        Asset search API.

        Possible kwargs:
           * threshold: int (default = 0)
           * page_no: int (default = 1)
           * page_size: int (default = 100)
           * sort_fields: list (default = [])
           * asset_types: list (default = [])
           * include_public: bool (default = True)
           * include_data_sources: bool (default = True)
        """
        self.logger.info('Asset Search - Asset Manager: %s', asset_manager_id)
        url = '{endpoint}/assets/search/{asset_manager_id}'.format(
            asset_manager_id=asset_manager_id,
            endpoint=self.endpoint,
        )
        params = {'query': query}
        for k, v in kwargs.items():
            if not isinstance(v, str) and isinstance(v, Iterable):
                v = ','.join(str(i) for i in v)

            params[k] = v

        response = self.session.get(url, params=params)
        if response.ok:
            data = response.json()
            assets = [json_to_asset(json_asset) for json_asset in data.get('hits', [])]
            self.logger.info('Returned %s Assets.', len(assets))
            return assets
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def search(self, asset_manager_id, **kwargs):
        """
            Search for assets.

            Possible kwargs:
             * client_ids: list[int]
             * asset_statuses: list
             * asset_ids: list
             * reference_types: list
             * reference_values: list
             * asset_issuer_ids: list[int]
             * asset_classes: list
             * asset_types: list
             * country_ids: list
             * currencies: list

             * include_public: bool (default = True)
             * include_data_sources: bool (default = True)

             * page_no = int(query_params.get('page_no', '1')) if query_params else None
             * page_size = int(query_params.get('page_size', '100')) if query_params else None

        """
        self.logger.info('Search for Assets - Asset Manager: %s', asset_manager_id)
        search_params = {}
        for k, v in kwargs.items():
            if not isinstance(v, str) and isinstance(v, Iterable):
                v = ','.join(str(i) for i in v)

            search_params[k] = v

        url = '%s/assets/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url, params=search_params)
        if response.ok:
            assets = [json_to_asset(json_asset) for json_asset in response.json()]
            self.logger.info('Returned %s Assets.', len(assets))
            return assets
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def fields_search(self, asset_manager_id, asset_ids=None,
                      asset_classes=None, asset_types=None,
                      fields=None, page_no=None, page_size=None):
        self.logger.info('Search for Assets - Asset Manager: %s', asset_manager_id)
        search_params = {}

        if asset_ids:
            search_params['asset_ids'] = ','.join(asset_ids)
        if asset_classes:
            search_params['asset_classes'] = ','.join(asset_classes)
        if asset_types:
            search_params['asset_types'] = ','.join(asset_types)
        if fields:
            search_params['fields'] = ','.join(fields)
        if page_no is not None:
            search_params['page_no'] = page_no
        if page_size:
            search_params['page_size'] = page_size

        url = '%s/assets/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url, params=search_params)
        if response.ok:
            asset_dicts = response.json()
            self.logger.info('Returned %s Assets.', len(asset_dicts))
            return asset_dicts
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def assets_by_asset_manager(self, asset_manager_id):
        self.logger.info('Retrieve Assets By Asset Manager: %s', asset_manager_id)
        url = '%s/assets/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url)
        if response.ok:
            assets = [json_to_asset(json_asset) for json_asset in response.json()]
            self.logger.info('Returned %s Assets.', len(assets))
            return assets
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def assets_lifecycle(self, asset_manager_id, business_date, asset_ids):
        self.logger.info('Retrieve Assets Lifecycle. Asset Manager: %s', asset_manager_id)
        url = '%s/asset-lifecycle/%s' % (self.endpoint, asset_manager_id)
        params = {'business_date': business_date.isoformat(),
                  'asset_ids': ','.join(asset_ids)}
        response = self.session.get(url, params=params)
        if response.ok:
            asset_lifecycles = response.json()
            self.logger.info('Returned %s Asset Lifecycles.', len(asset_lifecycles))
            return asset_lifecycles
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def clear(self, asset_manager_id):
        """ This method deletes all the data for an asset_manager_id.
            It should be used with extreme caution.  In production it
            is almost always better to Inactivate rather than delete. """
        self.logger.info('Clear Assets - Asset Manager: %s', asset_manager_id)
        url = '%s/clear/%s' % (self.endpoint, asset_manager_id)
        response = self.session.delete(url)
        if response.ok:
            count = response.json().get('count', 'Unknown')
            self.logger.info('Deleted %s Assets.', count)
            return count
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def synch(self, asset_manager_id, **params):
        """
        This method invokes a request to synch the assets in
        dynamodb and/or elastic search with the SQL assets.

        Args:
            asset_manager_id (int): The id of the asset manager that owns the assets to be synched
            cache (bool): Whether to synch the assets in DynamoDB
            search (bool): Whether to synch the assets in Elasticsearch
            page_no (int): The current page of assets to be synched.
            page_size (int): The number of assets to be synched per request.
        
        Returns:
            int: the number of assets synched.
        """
        self.logger.info('Synching Assets.')
        url = '%s/synch/%s' % (self.endpoint, asset_manager_id)
        response = self.session.put(url, params=params)

        if response.ok:
            count = response.json().get('count', 0)
            self.logger.info('Synched %s Assets.', count)
            return count
        else:
            self.logger.error(response.text)
            response.raise_for_status()


import csv

from amaascore.assets.interface import AssetsInterface
from amaascore.parties.interface import PartiesInterface
from amaascore.books.interface import BooksInterface
from amaascore.corporate_actions.interface import CorporateActionsInterface
from amaascore.market_data.interface import MarketDataInterface
from amaascore.transactions.interface import TransactionsInterface
from amaascore.asset_managers.interface import AssetManagersInterface
from amaascore import core
from amaascore import parties
from amaascore import assets
from amaascore import transactions


CHILDREN_CLASS = {'asset': {'links': assets.children.Link, 'references': core.reference.Reference},
                  'party': {'addresses': parties.children.Address, 'emails': parties.children.Email,
                            'links': parties.children.Link, 'references': core.reference.Reference,
                            'comments': core.comment.Comment},
                  'transaction': {'charges': transactions.children.Charge, 'codes': transactions.children.Code,
                                  'comments': core.comment.Comment, 'links': transactions.children.Link,
                                  'rates': transactions.children.Rate, 'parties': transactions.children.Party,
                                  'references': core.reference.Reference},
                  'asset_manager': {},
                  'book': {},
                  'corporate_action': {'references': core.reference.Reference}}

def direct_to_class(amaasclass):
    """direct from amaasclass (first params given in the row) to the dictionary of the children class"""
    if amaasclass in ASSET:
        return CHILDREN_CLASS['asset']
    elif amaasclass in PARTY:
        return CHILDREN_CLASS['party']
    elif amaasclass in TRANSACTION:
        return CHILDREN_CLASS['transaction']
    elif amaasclass in BOOK:
        return CHILDREN_CLASS['book']
    elif amaasclass in CORPORATE_ACTION:
        return CHILDREN_CLASS['corporate_action']
    elif amaasclass in MARKET_DATA:
        return None #None for now
    else:
        return CHILDREN_CLASS['asset_manager']

def interface_direct_class(data_class):
    """help to direct to the correct interface interacting with DB by class name only"""
    if data_class in ASSET:
        interface = AssetsInterface()
    elif data_class in PARTY:
        interface = PartiesInterface()
    elif data_class in BOOK:
        interface = BooksInterface()
    elif data_class in CORPORATE_ACTION:
        interface = CorporateActionsInterface()
    elif data_class in MARKET_DATA:
        interface = MarketDataInterface()
    elif data_class in TRANSACTION:
        interface = TransactionsInterface()
    else:
        interface = AssetManagersInterface()
    return interface

def interface_direct_csvpath(csvpath):
    """help to direct to the correct interface interacting with DB by csvfile path"""
    with open(csvpath) as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            data_class = row.pop('amaasclass', '')
            return interface_direct_class(data_class)

def process_normal(_dict):
    """
    this method process the _dict to correct dict to be called by class constructor
    this method will be imported and called by main csv uploader function
    """
    cooked_dict = group_raw_to_formatted_string_dict(_dict)
    data_class = cooked_dict.pop('amaasclass', '')
    children_class_dict = direct_to_class(data_class)
    tasty_dict = dict()
    for cooked_key, cooked_value in cooked_dict.items():
        if cooked_key in CHILDREN_SIGNAL:
            processed_dict = {cooked_key: formatted_string_to_others(cooked_value, children_class_dict[cooked_key])}
        elif cooked_key == 'links':
            processed_dict = {cooked_key: formatted_string_to_links(cooked_value, children_class_dict[cooked_key])}
        else:
            processed_dict = {cooked_key: process_value_with_header(cooked_key, cooked_value)}
        tasty_dict.update(processed_dict)
    return tasty_dict



import json
from decimal import Decimal
from amaascore.core.amaas_model import AMaaSModel

class PnlAmount(AMaaSModel):


    def __init__(self, total_pnl=None, asset_pnl=None, fx_pnl=None,
                 realised_pnl=None, unrealised_pnl=None,
                 additional=None, error_message=None):
        self.total_pnl = total_pnl
        self.asset_pnl = asset_pnl
        self.fx_pnl = fx_pnl
        self.realised_pnl = realised_pnl
        self.unrealised_pnl = unrealised_pnl
        self.additional = additional
        self.error_message = error_message
    
    @property
    def asset_pnl(self):
        return self._asset_pnl

    @asset_pnl.setter
    def asset_pnl(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._asset_pnl = Decimal(val)
        else:
            self._asset_pnl = val

    @property
    def fx_pnl(self):
        return self._fx_pnl

    @fx_pnl.setter
    def fx_pnl(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._fx_pnl = Decimal(val)
        else:
            self._fx_pnl = val

    @property
    def total_pnl(self):
        return self._total_pnl

    @total_pnl.setter
    def total_pnl(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._total_pnl = Decimal(val)
        else:
            self._total_pnl = val

    @property
    def unrealised_pnl(self):
        return self._unrealised_pnl

    @unrealised_pnl.setter
    def unrealised_pnl(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._unrealised_pnl = Decimal(val)
        else:
            self._unrealised_pnl = val

    @property
    def realised_pnl(self):
        return self._realised_pnl

    @realised_pnl.setter
    def realised_pnl(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._realised_pnl = Decimal(val)
        else:
            self._realised_pnl = val

    @property
    def additional(self):
        if hasattr(self, '_additional'):
            return self._additional

    @additional.setter
    def additional(self, val):
        if val:
            if isinstance(val, str):
                self._additional = json.loads(val)
            elif isinstance(val, dict):
                self._additional = val
            else:
                raise TypeError('Unsupported data type for additional.')


from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime
import logging

import requests
from warrant.aws_srp import AWSSRP

from amaascore.config import ENDPOINTS, ConfigFactory, get_factory
from amaascore.exceptions import AMaaSException


class AMaaSSession(object):

    def __init__(self, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.refresh_period = 45 * 60  # minutes * seconds
        self.last_authenticated = None
        self.session = requests.Session()

    def needs_refresh(self):
        if not (self.last_authenticated and
                (datetime.utcnow() - self.last_authenticated).seconds < self.refresh_period):
            return True
        else:
            return False

    def put(self, url, data=None, **kwargs):
        # Add a refresh
        if self.last_authenticated and not self.needs_refresh():
            return self.session.put(url=url, data=data, **kwargs)
        else:
            raise AMaaSException('Not Authenticated')

    def post(self, url, data=None, **kwargs):
        # Add a refresh
        if self.last_authenticated and not self.needs_refresh():
            return self.session.post(url=url, data=data, **kwargs)
        else:
            raise AMaaSException('Not Authenticated')

    def delete(self, url, **kwargs):
        # Add a refresh
        if self.last_authenticated and not self.needs_refresh():
            return self.session.delete(url=url, **kwargs)
        else:
            raise AMaaSException('Not Authenticated')

    def get(self, url, **kwargs):
        # Add a refresh
        if self.last_authenticated and not self.needs_refresh():
            return self.session.get(url=url, **kwargs)
        else:
            raise AMaaSException('Not Authenticated')

    def patch(self, url, data=None, **kwargs):
        # Add a refresh
        if self.last_authenticated and not self.needs_refresh():
            return self.session.patch(url=url, data=data, **kwargs)
        else:
            raise AMaaSException('Not Authenticated')


class AMaaSPasswordSession(AMaaSSession):

    __session_cache = {}

    def __new__(cls, config, username=None, password=None, logger=None):
        cache_key = (config, username, password)
        cached = cls.__session_cache.get(cache_key)
        if not cached:
            cached = super(AMaaSPasswordSession, cls).__new__(cls)
            cls.__session_cache[cache_key] = cached

        return cached

    def __init__(self, config, username=None, password=None, logger=None):
        super(AMaaSPasswordSession, self).__init__(logger)
        self.username = username or config.username
        self.password = password or config.password
        self.aws = AWSSRP(
            username=self.username, password=self.password,
            pool_id=config.cognito_pool_id,
            pool_region=config.cognito_region,
            client_id=config.cognito_client_id,
        )

        if self.needs_refresh():
            self.login()

    def login(self):
        try:
            self.logger.info("Attempting login for: %s", self.username)
            tokens = self.aws.authenticate_user().get('AuthenticationResult')
            self.logger.info("Login successful")
            self.last_authenticated = datetime.utcnow()
            self.session.headers.update({'Authorization': tokens.get('IdToken')})
        except self.aws.client.exceptions.NotAuthorizedException:
            self.logger.exception("Login failed")
            self.last_authenticated = None


class AMaaSTokenSession(AMaaSSession):

    def __init__(self, session_token, logger=None):
        super(AMaaSTokenSession, self).__init__(logger)
        self.session_token = session_token
        self.logger.info("Skipping login since session token is provided.")
        self.session.headers.update({'Authorization': self.session_token})
        self.last_authenticated = datetime.utcnow()


class Interface(object):
    """
    Currently this class doesn't do anything - but I anticipate it will be needed in the future.
    """

    def __init__(self, endpoint_type, endpoint=None, environment=None, username=None, password=None,
                 config_filename=None, logger=None, session_token=None):
        self.logger = logger or logging.getLogger(__name__)
        if config_filename:
            config_factory = ConfigFactory(config_filename)
        else:
            config_factory = get_factory()

        self.api_config = config_factory.api_config(environment)

        self.endpoint_type = endpoint_type
        self.endpoint = self.get_endpoint(endpoint)
        if session_token:
            self.session = AMaaSTokenSession(session_token, logger=self.logger)
        else:
            self.session = AMaaSPasswordSession(
                config_factory.auth_config(environment),
                username=username, password=password,
                logger=self.logger,
            )

        self.json_header = {'Content-Type': 'application/json'}
        self.logger.info('Interface Created')

    def get_endpoint(self, endpoint=None):
        """Return interface URL endpoint."""
        base_url = self.api_config.api_url
        if not endpoint:
            if 'localhost' in base_url:
                endpoint = ''
            else:
                endpoint = ENDPOINTS[self.endpoint_type]

        endpoint = '/'.join([p.strip('/') for p in (base_url, endpoint)]).strip('/')
        self.logger.info("Using Endpoint: %s", endpoint)
        return endpoint



import pytz
from copy import copy
from datetime import datetime, date
from decimal import Decimal
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta
from amaascore.core.amaas_model import AMaaSModel
from amaascore.transactions.pnl_amount import PnlAmount


class Pnl(AMaaSModel):

    def __init__(self, asset_manager_id, pnl_type,
                 book_id, business_date, pnl_timestamp,
                 currency, quantity=None, asset_id=None,
                 transaction_id=None, transaction_date=None,
                 DTD=None, MTD=None, YTD=None,
                 *args, **kwargs):
        self.asset_manager_id = asset_manager_id
        self.pnl_type = pnl_type
        self.book_id = book_id
        self.business_date = business_date
        self.pnl_timestamp = pnl_timestamp
        self.currency = currency
        self.quantity = quantity
        self.asset_id = asset_id
        self.transaction_id = transaction_id
        self.transaction_date = transaction_date
        self.DTD = copy(DTD)
        self.MTD = copy(MTD)
        self.YTD = copy(YTD)

    @property
    def quantity(self):
        return self._quantity

    @quantity.setter
    def quantity(self, value):
        if not isinstance(value, Decimal) and value is not None:
            self._quantity = Decimal(value)
        else:
            self._quantity = value

    @property
    def asset_manager_id(self):
        if hasattr(self, '_asset_manager_id'):
            return self._asset_manager_id

    @asset_manager_id.setter
    def asset_manager_id(self, value):
        self._asset_manager_id = int(value)

    @property
    def DTD(self):
        return self._DTD

    @DTD.setter
    def DTD(self, value):
        if value is None:
            self._DTD = None
        elif isinstance(value, dict):
            self._DTD = PnlAmount(**value)
        elif isinstance(value, PnlAmount):
            self._DTD = value
        else:
            raise TypeError('Invalid type for DTD')

    @property
    def MTD(self):
        return self._MTD

    @MTD.setter
    def MTD(self, value):
        if value is None:
            self._MTD = None
        elif isinstance(value, dict):
            self._MTD = PnlAmount(**value)
        elif isinstance(value, PnlAmount):
            self._MTD = value
        else:
            raise TypeError('Invalid type for MTD')

    @property
    def YTD(self):
        return self._YTD

    @YTD.setter
    def YTD(self, value):
        if value is None:
            self._YTD = None
        elif isinstance(value, dict):
            self._YTD = PnlAmount(**value)
        elif isinstance(value, PnlAmount):
            self._YTD = value
        else:
            raise TypeError('Invalid type for YTD')

    @property
    def pnl_type(self):
        return self._pnl_type

    @pnl_type.setter
    def pnl_type(self, value):
        if value not in ['Position', 'Transaction']:
            raise ValueError('Invalid value for "Pnl Type"')
        self._pnl_type = value

    @property
    def business_date(self):
        if hasattr(self, '_business_date'):
            return self._business_date

    @business_date.setter
    def business_date(self, value):
        if value is None:
            raise ValueError('Missing required attribute "Business Date".')

        if isinstance(value, str):
            self._business_date = parse(value).date()
        elif isinstance(value, datetime):
            self._business_date = value.date()
        elif isinstance(value, date):
            self._business_date = value
        else:
            raise TypeError('Invalid type for attribute "Business Date".')

    @property
    def pnl_timestamp(self):
        return self._pnl_timestamp

    @pnl_timestamp.setter
    def pnl_timestamp(self, value):
        if value is None:
            raise ValueError('Missing required attribute "Pnl Timestamp".')

        parsed_value = None
        if isinstance(value, str):
            parsed_value = parse(value)
        elif isinstance(value, datetime):
            parsed_value = value
        elif isinstance(value, date):
            parsed_value = datetime.combine(value, datetime.min.time())
        else:
            raise TypeError('Invalid type for attribute "Pnl Timestamp".')

        if parsed_value is None:
            raise ValueError('Unable to parse {} into a "Pnl Timestamp".'.format(str(value)))
        else:
            self._pnl_timestamp = parsed_value.replace(tzinfo=pytz.UTC) \
                                  if not parsed_value.tzinfo else parsed_value

    @property
    def transaction_date(self):
        if hasattr(self, '_transaction_date'):
            return self._transaction_date

    @transaction_date.setter
    def transaction_date(self, value):
        if value is None:
            raise ValueError('Missing required attribute "Transaction Date".')

        if isinstance(value, str):
            self._transaction_date = parse(value).date()
        elif isinstance(value, datetime):
            self._transaction_date = value.date()
        elif isinstance(value, date):
            self._transaction_date = value
        else:
            raise TypeError('Invalid type for attribute "Transaction Date".')


from __future__ import absolute_import, division, print_function, unicode_literals

import json
import logging

from amaascore.core.amaas_model import json_handler
from amaascore.core.interface import Interface
from amaascore.market_data.utils import json_to_eod_price, json_to_fx_rate, json_to_curve


class MarketDataInterface(Interface):

    def __init__(self, environment=None, logger=None, endpoint=None, username=None, 
                 password=None, session_token=None):
        self.logger = logger or logging.getLogger(__name__)
        super(MarketDataInterface, self).__init__(endpoint=endpoint, endpoint_type='market_data', session_token=session_token,
                                                  environment=environment, username=username, password=password)

    def persist_eod_prices(self, asset_manager_id, business_date, eod_prices, update_existing_prices=True):
        """

        :param asset_manager_id:
        :param business_date: The business date for which these are rates.  Not really needed, could be derived...
        :param eod_prices:
        :param update_existing_prices:
        :return:
        """
        self.logger.info('Persist EOD Prices - Asset Manager: %s - Business Date: %s', asset_manager_id, business_date)
        url = '%s/eod-prices/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        params = {'update_existing_prices': update_existing_prices}
        eod_prices_json = [eod_price.to_interface() for eod_price in eod_prices]
        response = self.session.post(url, params=params, json=eod_prices_json)
        if response.ok:
            eod_prices = [json_to_eod_price(eod_price) for eod_price in response.json()]
            return eod_prices
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def retrieve_eod_prices(self, asset_manager_id, business_date, asset_ids=None, last_available=False):
        self.logger.info('Retrieve EOD Prices - Asset Manager: %s - Business Date: %s', asset_manager_id, business_date)
        url = '%s/eod-prices/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        params = {'asset_ids': ','.join(asset_ids)} if asset_ids else {}
        # setting last_available to True will return the latest price of the asset as of business
        # date (or latest prior price if price found on the date)
        if last_available:
            params.update({'last_available':True})
        response = self.session.get(url=url, params=params)
        if response.ok:
            eod_prices = [json_to_eod_price(eod_price) for eod_price in response.json()]
            self.logger.info('Returned %s EOD Prices.', len(eod_prices))
            return eod_prices
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def roll_prices(self, asset_manager_id, previous_date, asset_ids, update_existing_prices=False):
        url = '%s/roll-prices/%s' % (self.endpoint, asset_manager_id)
        params = {'update_existing_prices': update_existing_prices}
        json_body = json.loads(json.dumps({'previous_date': previous_date,
                                           'asset_ids': ','.join(asset_ids)}, default=json_handler))
        response = self.session.post(url=url, params=params, json=json_body)
        if response.ok:
            prices = [json_to_eod_price(price) for price in response.json()]
            self.logger.info('Rolled %s Prices.', len(prices))
            return prices
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def persist_fx_rates(self, asset_manager_id, business_date, fx_rates, update_existing_rates=True):
        """

        :param asset_manager_id:
        :param business_date: The business date for which these are rates.  Not really needed, could be derived...
        :param fx_rates:
        :param update_existing_rates:
        :return:
        """
        self.logger.info('Persist FX Rates - Asset Manager: %s - Business Date: %s', asset_manager_id, business_date)
        url = '%s/fx-rates/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        params = {'update_existing_rates': update_existing_rates}
        fx_rates_json = [fx_rate.to_interface() for fx_rate in fx_rates]
        response = self.session.post(url, params=params, json=fx_rates_json)
        if response.ok:
            fx_rates = [json_to_fx_rate(fx_rate) for fx_rate in response.json()]
            return fx_rates
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def retrieve_fx_rates(self, asset_manager_id, business_date, asset_ids=None):
        self.logger.info('Retrieve FX Rates - Asset Manager: %s - Business Date: %s', asset_manager_id, business_date)
        url = '%s/fx-rates/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        params = {'asset_ids': ','.join(asset_ids)} if asset_ids else {}
        response = self.session.get(url=url, params=params)
        if response.ok:
            fx_rates = [json_to_fx_rate(fx_rate) for fx_rate in response.json()]
            self.logger.info('Returned %s FX Rates.', len(fx_rates))
            return fx_rates
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def search_fx_rates(self, asset_manager_id, client_ids=None,
                        asset_ids=None, rate_types=None, rate_statuses=None,
                        business_date_start=None, business_date_end=None,
                        rate_timestamp_start=None, rate_timestamp_end=None):
        self.logger.info('Search FX Rates - Asset Manager: %s', asset_manager_id)
        url = '%s/fx-rates/%s' % (self.endpoint, asset_manager_id)

        params = {}
        if client_ids:
            params['client_ids'] = ','.join(client_ids)
        if asset_ids:
            params['asset_ids'] = ','.join(asset_ids)
        if rate_types:
            params['rate_types'] = ','.join(rate_types)
        if rate_statuses:
            params['rate_statuses'] = ','.join(rate_statuses)
        if business_date_start:
            params['business_date_start'] = business_date_start.date().isoformat()
        if business_date_end:
            params['business_date_end'] = business_date_end.date().isoformat()
        if rate_timestamp_start:
            params['rate_timestamp_start'] = rate_timestamp_start.strftime('%H:%m:%s')
        if rate_timestamp_end:
            params['rate_timestamp_end'] = rate_timestamp_end.strftime('%H:%m:%s')

        response = self.session.get(url=url, params=params)
        if response.ok:
            fx_rates = [json_to_fx_rate(fx_rate) for fx_rate in response.json()]
            self.logger.info('Returned %s FX Rates.', len(fx_rates))
            return fx_rates
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_curve(self, asset_manager_id, business_date, asset_ids = None):
        self.logger.info('Retrieve curve - Asset Manager: %s - Business Date: %s', asset_manager_id, business_date)
        url = '%s/curves/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        params = {'asset_ids': ','.join(asset_ids)} if asset_ids else {}
        response = self.session.get(url = url, params = params)
        if response.ok:
            curves = [json_to_curve(curve) for curve in response.json() ]
            self.logger.info('Returned %s curves.', len(curves))
            return curves
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def persist_curves(self, asset_manager_id, business_date, curves, update_existing_curves=True):
        """
        :param asset_manager_id:
        :param business_date:
        :param curves:
        :param update_existing_rates:
        :return:
        """
        self.logger.info('Persist curves - Asset Manager: %s - Business Date: %s', asset_manager_id, business_date)
        url = '%s/curves/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        params = {'update_existing_curves': update_existing_curves}
        curves_json = [curve.to_interface() for curve in curves]
        response = self.session.post(url, params = params, json=curves_json)
        if response.ok:
            curves = [json_to_curve(curve) for curve in response.json()]
            return curves
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def clear(self, asset_manager_id):
        """ This method deletes all the data for an asset_manager_id.
            It should be used with extreme caution.  In production it
            is almost always better to Inactivate rather than delete. """
        self.logger.info('Clear Market Data - Asset Manager: %s', asset_manager_id)
        url = '%s/clear/%s' % (self.endpoint, asset_manager_id)
        response = self.session.delete(url)
        if response.ok:
            eod_price_count = response.json().get('eod_price_count', 'Unknown')
            self.logger.info('Deleted %s EOD Prices.', eod_price_count)
            fx_rate_count = response.json().get('fx_rate_count', 'Unknown')
            self.logger.info('Deleted %s FX Rates.', fx_rate_count)
            return response.json()
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def get_brokendate_fx_forward_rate(self, asset_manager_id,  asset_id, price_date, value_date):
        """
        This method takes calculates broken date forward FX rate based on the passed in parameters
        """
        self.logger.info('Calculate broken date FX Forward - Asset Manager: %s - Asset (currency): %s - Price Date: %s - Value Date: %s', asset_manager_id, asset_id, price_date, value_date)
        url = '%s/brokendateforward/%s' % (self.endpoint, asset_manager_id)
        params = {'value_date': value_date, 'asset_id':asset_id, 'price_date': price_date}
        response = self.session.get(url=url, params = params)
        if response.ok:
            forward_rate = response.json()
            self.logger.info('Retrieved broken date FX forward rate %s - %s: %s', asset_id, price_date, value_date)
            return forward_rate
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def market_data_retrieve(self, asset_manager_id, market_data_set):
        self.logger.info('Retrieving custom market data set')
        url = '%s/market-data/%s' % (self.endpoint, asset_manager_id)
        response = self.session.post(url, json=market_data_set)
        if response.ok:
            self.logger.info('Recieved %s custom market data points', str(len(response.json())))
            return response.json()
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def last_available_business_date(self, asset_manager_id, asset_ids, page_no=None, page_size=None):
        """
        Returns the last available business date for the assets so we know the 
        starting date for new data which needs to be downloaded from data providers.
        
        This method can only be invoked by system user
        """
        self.logger.info('Retrieving last available business dates for assets')
        url = '%s/last-available-business-date' % self.endpoint
        params = {'asset_manager_ids': [asset_manager_id],
                  'asset_ids': ','.join(asset_ids)}
        if page_no:
            params['page_no'] = page_no
        if page_size:
            params['page_size'] = page_size
        response = self.session.get(url, params=params)
        if response.ok:
            self.logger.info("Received %s assets' last available business date", len(response.json()))
            return response.json()
        else:
            self.logger.error(response.text)
            response.raise_for_status()


from decimal import Decimal
import json
from amaascore.core.amaas_model import AMaaSModel


class BookReport(AMaaSModel):

    def __init__(self, asset_manager_id, book_id, period, business_date, report_timestamp,
                 report_status='Active', currency=None, total_pnl=None, asset_pnl=None, fx_pnl=None, 
                 mtm_value=None, report_type=None, message=None, *args, **kwargs):
        
        self.asset_manager_id = asset_manager_id
        self.book_id = book_id
        self.period = period
        self.business_date = business_date
        self.currency = currency
        self.total_pnl = total_pnl
        self.asset_pnl = asset_pnl
        self.fx_pnl = fx_pnl
        self.report_status = report_status
        self.message = message
        self.mtm_value = mtm_value
        self.report_timestamp = report_timestamp
        self.report_type = report_type

        super(BookReport, self).__init__(*args, **kwargs)

    @property
    def report_type(self):
        return self._report_type

    @report_type.setter
    def report_type(self, val):
        if val not in ['MTM', 'PNL']:
            raise ValueError("""Unrecognized book report type: %s,
                            expect either 'MTM' or 'PNL'""" % str(val))
        self._report_type = val

    @property
    def period(self):
        return self._period

    @period.setter
    def period(self, val):
        if val not in ['YTD', 'MTD', 'DTD', 'N/A']:
            raise ValueError("""Unrecognized PnL period %s, expect
                    period to be one of the following: 'YTD', 'MTD', 'DTD', 'N/A'""" % str(val))
        self._period = val

    @property
    def total_pnl(self):
        return self._total_pnl

    @total_pnl.setter
    def total_pnl(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._total_pnl = Decimal(val)
        else:
            self._total_pnl = val

    @property
    def fx_pnl(self):
        return self._fx_pnl

    @fx_pnl.setter
    def fx_pnl(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._fx_pnl = Decimal(val)
        else:
            self._fx_pnl = val

    @property
    def asset_pnl(self):
        return self._asset_pnl

    @asset_pnl.setter
    def asset_pnl(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._asset_pnl = Decimal(val)
        else:
            self._asset_pnl = val

    @property
    def mtm_value(self):
        return self._mtm_value

    @mtm_value.setter
    def mtm_value(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._mtm_value = Decimal(val)
        else:
            self._mtm_value = val


from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.core.amaas_model import AMaaSModel


class Reference(AMaaSModel):

    def __init__(self, reference_value, reference_primary=False, *args, **kwargs):
        self.reference_value = reference_value
        self.reference_primary = reference_primary
        super(Reference, self).__init__(*args, **kwargs)

    @property
    def reference_primary(self):
        if hasattr(self, '_reference_primary'):
            return self._reference_primary

    @reference_primary.setter
    def reference_primary(self, value):
        """
        Always convert to bool if the service/database returns 0 or 1
        """
        if value is not None:
            self._reference_primary = True if value else False


"""Exception classes."""


class AMaaSException(Exception):
    """Base class for all AMaaS Exceptions."""


class TransactionNeedsSaving(AMaaSException):
    def __init__(self):
        message = "Transaction needs to be saved to AMaaS Core for the functionality to be valid"
        super(TransactionNeedsSaving, self).__init__(message)


class ConfigurationError(AMaaSException):
    """Configuration error."""


class MissingConfigurationError(ConfigurationError):
    """Required configuration is missing."""

    def __init__(self, section, name):
        msg = 'Configuration missing: {section}->{name}'
        super(ConfigurationError, self).__init__(msg.format(section=section, name=name))



import logging.config

from amaascore.tools.csv_tools import csv_stream_to_objects
from amaasutils.logging_utils import DEFAULT_LOGGING
from amaascore.csv_upload.utils import process_normal, interface_direct_csvpath


class Uploader(object):

    @staticmethod
    def json_handler(orderedDict, params):
        Dict = dict(orderedDict)
        for key, var in params.items():
            Dict[key]=var
        data_class = Dict.get('amaasclass', None)
        Dict = process_normal(Dict)
        obj = globals()[data_class](**dict(Dict))
        return obj

    @staticmethod
    def upload(csvpath, asset_manager_id=None, client_id=None):
        """convert csv file rows to objects and insert;
           asset_manager_id and possibly client_id from the UI (login)"""
        interface = interface_direct_csvpath(csvpath)
        logging.config.dictConfig(DEFAULT_LOGGING)
        logger = logging.getLogger(__name__)
        if asset_manager_id is None:
            params = dict()
        elif client_id is None:
            params = {'asset_manager_id': asset_manager_id}
        else:
            params = {'asset_manager_id': asset_manager_id, 'client_id': client_id}
        with open(csvpath) as csvfile:
            objs = csv_stream_to_objects(stream=csvfile, json_handler=Uploader.json_handler, params=params)
        for obj in objs:
            interface.new(obj)
            logger.info('Creating this object and upload to database successfully')

    @staticmethod
    def download(csvpath, asset_manager_id, data_id_type, data_id_list):
        """retrieve the objs mainly for test purposes"""
        interface = interface_direct_csvpath(csvpath)
        logging.config.dictConfig(DEFAULT_LOGGING)
        logger = logging.getLogger(__name__)
        objs = []
        for data_id in data_id_list:
            Dict = dict()
            Dict[data_id_type] = data_id
            objs.append(interface.retrieve(asset_manager_id=asset_manager_id, **Dict))
        return objs
